from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

import attrs
import graphql
from attr import define

from qtgqlcodegen.utils import HashAbleDict

if TYPE_CHECKING:
    from graphql.language import ast as gql_lang
    from graphql.type import definition as gql_def

    from qtgqlcodegen.schema.definitions import (
        QtGqlArgumentDefinition,
        QtGqlFieldDefinition,
        QtGqlVariableDefinition,
        SchemaTypeInfo,
    )
    from qtgqlcodegen.types import (
        QtGqlObjectType,
        QtGqlQueriedInterface,
        QtGqlQueriedObjectType,
        QtGqlTypeABC,
    )


@attrs.define(slots=False, repr=False)
class OperationTypeInfo:
    schema_type_info: SchemaTypeInfo
    narrowed_types_map: dict[str, QtGqlQueriedObjectType] = attrs.Factory(dict)
    narrowed_interfaces_map: dict[str, QtGqlQueriedInterface] = attrs.Factory(dict)
    variables: list[QtGqlVariableDefinition] = attrs.Factory(list)
    used_fragments: HashAbleDict[str, gql_lang.FragmentDefinitionNode] = attrs.Factory(HashAbleDict)
    available_fragments: HashAbleDict[str, gql_lang.FragmentDefinitionNode] = attrs.Factory(
        HashAbleDict,
    )


@attrs.define(frozen=True, slots=False, repr=False)
class QtGqlQueriedField:
    type: QtGqlTypeABC
    type_info: OperationTypeInfo
    origin: QtGqlObjectType
    concrete: QtGqlFieldDefinition
    variable_uses: list[QtGqlVariableUse] = attrs.Factory(list)

    @cached_property
    def cached_by_args(self) -> bool:
        # if the origin implements node it's fields are cached by arguments
        # if they have ones
        return bool(self.concrete.arguments)

    @cached_property
    def is_root(self) -> bool:
        return self.origin.name in self.type_info.schema_type_info.root_types_names

    @cached_property
    def type_name(self) -> str:
        return self.type.member_type

    @cached_property
    def build_variables_tuple_for_field_arguments(self) -> str:
        # operation might not use an argument that has default value, ignore what's ignored.
        # see https://github.com/qtgql/qtgql/issues/272 for more details.
        if self.cached_by_args and self.variable_uses:
            assert len(self.concrete.arguments) == len(self.variable_uses)
            return (
                "{"
                + ",".join(
                    [
                        f"operation->vars_inst.{arg.variable.name}.value()"
                        for arg in self.variable_uses
                    ],
                )
                + "}"
            )
        return ""

    @property
    def name(self) -> str:
        return self.concrete.name

    @property
    def private_name(self):
        return self.concrete.private_name


@define(slots=False, repr=False)
class QtGqlOperationDefinition:
    operation_def: gql_def.OperationDefinitionNode
    root_type: QtGqlQueriedObjectType
    root_field: QtGqlQueriedField
    variables: list[QtGqlVariableDefinition] = attrs.Factory(list)
    narrowed_types: tuple[QtGqlQueriedObjectType, ...] = attrs.Factory(tuple)
    interfaces: tuple[QtGqlQueriedInterface, ...] = attrs.Factory(tuple)
    used_fragments: tuple[gql_lang.FragmentDefinitionNode, ...] = attrs.Factory(tuple)

    @property
    def generated_variables_type(self) -> str:
        return self.name + "Variables"

    @property
    def name(self) -> str:
        assert self.operation_def.name
        return self.operation_def.name.value

    @property
    def query(self) -> str:
        return "\n".join(
            [
                graphql.print_ast(self.operation_def),
                *[graphql.print_ast(frag) for frag in self.used_fragments],
            ],
        )


@define(slots=False, repr=False)
class QtGqlVariableUse:
    argument: tuple[int, QtGqlArgumentDefinition]
    variable: QtGqlVariableDefinition



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any

import jinja2
from attrs import define
from jinja2 import Environment, PackageLoader, select_autoescape

if TYPE_CHECKING:  # pragma: no cover
    from qtgqlcodegen.config import QtGqlConfig
    from qtgqlcodegen.generator import GenerationOutput

template_env: Environment = Environment(
    loader=PackageLoader("qtgqlcodegen"),
    autoescape=select_autoescape(),
    variable_start_string="👉",  # originally {{ variable }}, using 👉 variable 👈 because C++ uses curly brackets.
    variable_end_string="👈",
    undefined=jinja2.StrictUndefined,
)


def debug_jinja(obj: Any) -> None:  # pragma: no cover
    warnings.warn("jinja debug is called", stacklevel=2)
    return obj


template_env.globals.update(debug_jinja=debug_jinja)

CMAKE_TEMPLATE = template_env.get_template("CMakeLists.jinja.cmake")


@define(slots=False)
class CmakeTemplateContext:
    config: QtGqlConfig
    generation_output: GenerationOutput

    @property
    def target_name(self) -> str:
        return self.config.env_name


def cmake_template(context: CmakeTemplateContext) -> str:
    return CMAKE_TEMPLATE.render(context=context)



from functools import cached_property
from pathlib import Path

import graphql
from attrs import Factory, define

from qtgqlcodegen.generator import SchemaGenerator
from qtgqlcodegen.schema.definitions import CustomScalarMap
from qtgqlcodegen.types import CUSTOM_SCALARS


@define(slots=False)
class QtGqlConfig:
    """Encapsulates configurations for a qtgql-codegen application per GraphQL
    schema."""

    graphql_dir: Path
    """A directory contains.

    - schema.graphql, represents the current schema definition at the server.
    - operations.graphql, queries, mutations and subscription handlers would be generated based on the operations defined there.
    """
    env_name: str = "QGqlEnv"
    """The generated types would find the environment by this name.

    Also the generated QML imports would fall under this namespace.
    """
    custom_scalars: CustomScalarMap = Factory(dict)
    """Mapping of custom scalars, respected by the schema evaluator."""
    qml_plugins_path: str = "${CMAKE_BINARY_DIR}/qml"
    """Qml plugins would be installed under this directory.

    This should suffice for most basic setups.
    """
    debug: bool = False
    """Templates would render some additional helpers for testing."""

    @cached_property
    def schema_path(self) -> Path:
        return self.graphql_dir / "schema.graphql"

    @cached_property
    def operations_dir(self) -> Path:
        return self.graphql_dir / "operations.graphql"

    @cached_property
    def generated_dir(self):
        ret = self.graphql_dir / "__generated__"
        if not ret.exists():
            ret.mkdir()
        return ret

    @cached_property
    def _evaluator(self) -> SchemaGenerator:
        return SchemaGenerator(
            config=self,
            schema=graphql.build_schema(
                (self.graphql_dir / "schema.graphql").resolve(True).read_text("utf-8"),
            ),
        )

    def generate(self) -> None:
        self._evaluator.dump()

    def __attrs_post_init__(self):
        if self.custom_scalars != CUSTOM_SCALARS:
            self.custom_scalars.update(CUSTOM_SCALARS)



from __future__ import annotations

from typing import TYPE_CHECKING

from attr import define

from qtgqlcodegen.core.graphql_ref import (
    has_id_selection,
    has_typename_selection,
    inject_id_selection,
    inject_typename_selection,
    is_field_node,
    is_inline_fragment,
)
from qtgqlcodegen.utils import require

if TYPE_CHECKING:
    from graphql import language as gql_lang

    from qtgqlcodegen.schema.definitions import SchemaTypeInfo
    from qtgqlcodegen.types import (
        QtGqlInterface,
        QtGqlObjectType,
        QtGqlTypeABC,
        QtGqlUnion,
    )


@define
class SelectionsStatus:
    id_was_selected: bool = False
    typename_was_selected: bool = False
    """On interfaces inner inline fragments should not be injected if their
    parents was injected."""


def _resolve_union(
    type_info: SchemaTypeInfo,
    ss: gql_lang.SelectionSetNode,
    concrete: QtGqlUnion,
) -> None:
    selection_status = SelectionsStatus()
    if not has_typename_selection(ss):
        inject_typename_selection(ss)
        selection_status.typename_was_selected = True
    for selection in ss.selections:
        if is_field_node(selection):
            continue  # __typename selection
        fragment = is_inline_fragment(selection)
        assert fragment
        type_name = fragment.type_condition.name.value
        # unions support only object types http://spec.graphql.org/October2021/#sec-Unions
        _resolve_type(
            type_info,
            fragment.selection_set,
            require(concrete.get_by_name(type_name)),
        )


def _resolve_interface(
    type_info: SchemaTypeInfo,
    ss: gql_lang.SelectionSetNode,
    concrete: QtGqlInterface,
    selections_status: SelectionsStatus,
) -> None:
    if not selections_status.typename_was_selected and not has_typename_selection(ss):
        inject_typename_selection(ss)
        selections_status.typename_was_selected = True

    if (
        concrete.implements_node
        and not selections_status.id_was_selected
        and not has_id_selection(ss)
    ):
        inject_id_selection(ss)
        selections_status.id_was_selected = True

    for node in ss.selections:
        if inline_frag := is_inline_fragment(node):
            type_name = inline_frag.type_condition.name.value
            concrete_choice = type_info.get_object_or_interface(type_name)
            if inner_interface := concrete_choice.is_interface:
                _resolve_interface(
                    type_info,
                    inline_frag.selection_set,
                    inner_interface,
                    selections_status,
                )
            else:
                _resolve_object(
                    type_info,
                    inline_frag.selection_set,
                    require(concrete_choice.is_object_type),
                    selections_status.id_was_selected,
                )


def _resolve_object(
    type_info: SchemaTypeInfo,
    ss: gql_lang.SelectionSetNode,
    concrete: QtGqlObjectType,
    id_was_selected: bool = False,
) -> None:
    for selection in ss.selections:
        if f := is_field_node(selection):
            if f.selection_set:
                _resolve_type(type_info, f.selection_set, _get_field_type(concrete, f))
            if f.name.value == "id":
                id_was_selected = True

    if concrete.implements_node and not id_was_selected:
        inject_id_selection(ss)


def _get_field_type(
    parent: QtGqlObjectType | QtGqlInterface,
    field: gql_lang.FieldNode,
) -> QtGqlTypeABC:
    return parent.fields_dict[field.name.value].type


def _resolve_type(
    type_info: SchemaTypeInfo,
    ss: gql_lang.SelectionSetNode,
    tp: QtGqlTypeABC,
) -> None:
    if tp.is_model:
        tp = tp.is_model.of_type

    if tp.is_object_type:
        _resolve_object(type_info, ss, tp.is_object_type)
    elif tp.is_interface:
        _resolve_interface(type_info, ss, tp.is_interface, SelectionsStatus())
    elif tp.is_union:
        _resolve_union(type_info, ss, tp.is_union)


def inject_required_selections(
    type_info: SchemaTypeInfo,
    ss: gql_lang.SelectionSetNode,
    concrete: QtGqlObjectType | QtGqlInterface,
) -> None:
    """Qtgql relies on some selections based on some conditions.

    - If a type implements Node it should select `id`
    - If type is an interface it should select `__typename`
    """
    _resolve_type(type_info, ss, concrete)



from __future__ import annotations

from typing import NamedTuple

from graphql import OperationType
from graphql.type import definition as gql_def

from qtgqlcodegen.core.exceptions import QtGqlException
from qtgqlcodegen.core.graphql_ref import (
    is_enum_definition,
    is_input_definition,
    is_interface_definition,
    is_list_definition,
    is_non_null_definition,
    is_object_definition,
    is_scalar_definition,
    is_union_definition,
)
from qtgqlcodegen.schema.definitions import (
    CustomScalarMap,
    QtGqlArgumentDefinition,
    QtGqlFieldDefinition,
    SchemaTypeInfo,
)
from qtgqlcodegen.types import (
    BuiltinScalars,
    EnumValue,
    QtGqlDeferredType,
    QtGqlEnumDefinition,
    QtGqlInputObjectTypeDefinition,
    QtGqlInterface,
    QtGqlList,
    QtGqlObjectType,
    QtGqlOptional,
    QtGqlTypeABC,
    QtGqlUnion,
)


def evaluate_input_type(
    type_info: SchemaTypeInfo,
    type_: gql_def.GraphQLInputObjectType,
) -> QtGqlInputObjectTypeDefinition:
    ret = type_info.input_objects.get(type_.name, None)
    if not ret:
        ret = QtGqlInputObjectTypeDefinition(
            name=type_.name,
            docstring=type_.description,
            fields_dict={
                name: _evaluate_argument_field(type_info, name, field)
                for name, field in type_.fields.items()
            },
        )
        type_info.input_objects[ret.name] = ret
    return ret


def evaluate_graphql_type(
    type_info: SchemaTypeInfo,
    t: gql_def.GraphQLType,
) -> QtGqlTypeABC:
    # even though every type in qtgql has a default constructor,
    # hence there is no "real" non-null values
    # we store it as optional for generating nullable checks only for what's required.
    ret: QtGqlTypeABC | None = None
    is_optional = True
    if non_null := is_non_null_definition(t):
        t = non_null.of_type
        is_optional = False

    if list_def := is_list_definition(t):
        ret = QtGqlList(
            of_type=evaluate_graphql_type(type_info, list_def.of_type),
        )
    elif scalar_def := is_scalar_definition(t):
        if builtin_scalar := BuiltinScalars.by_graphql_name(scalar_def.name):
            ret = builtin_scalar
        else:
            ret = type_info.custom_scalars[scalar_def.name]

    elif enum_def := is_enum_definition(t):
        ret = _evaluate_enum(type_info, enum_def)

    elif obj_def := is_object_definition(t):
        ret = QtGqlDeferredType(
            object_map__=type_info.object_types,
            name=obj_def.name,
        )
    elif union_def := is_union_definition(t):
        ret = QtGqlUnion(
            types=tuple(
                QtGqlDeferredType(
                    name=possible.name,
                    object_map__=type_info.object_types,
                )
                for possible in type_info.schema_definition.get_possible_types(union_def)
            ),
        )
    elif input_def := is_input_definition(t):
        concrete = type_info.schema_definition.get_type(input_def.name)
        assert isinstance(concrete, gql_def.GraphQLInputObjectType)
        ret = evaluate_input_type(type_info, input_def)

    elif interface_def := is_interface_definition(t):
        ret = _evaluate_interface_type(type_info, interface_def)
    if not ret:  # pragma: no cover
        raise NotImplementedError(f"type {t} not supported yet")

    if is_optional:
        return QtGqlOptional(of_type=ret)
    return ret


def evaluate_field(
    type_info: SchemaTypeInfo,
    name: str,
    field: gql_def.GraphQLField,
) -> QtGqlFieldDefinition:
    return QtGqlFieldDefinition(
        type=evaluate_graphql_type(type_info, field.type),
        name=name,
        description=field.description,
        arguments_dict={
            name: _evaluate_argument_field(type_info, name, arg) for name, arg in field.args.items()
        },
    )


def _evaluate_argument_field(
    type_info: SchemaTypeInfo,
    name: str,
    field: gql_def.GraphQLInputField | gql_def.GraphQLArgument,
) -> QtGqlArgumentDefinition:
    return QtGqlArgumentDefinition(
        type=evaluate_graphql_type(type_info, field.type),
        name=name,
        description=field.description,
    )


class InterfaceOrObjectOptions(NamedTuple):
    implements: tuple[QtGqlInterface, ...]
    all_fields: dict[str, QtGqlFieldDefinition]
    unique_fields: tuple[QtGqlFieldDefinition, ...]


def _evaluate_object_fields(
    type_info: SchemaTypeInfo,
    obj: gql_def.GraphQLObjectType | gql_def.GraphQLInterfaceType,
) -> InterfaceOrObjectOptions:
    implements = tuple(
        _evaluate_interface_type(type_info, interface) for interface in obj.interfaces
    )
    inherited_fields = {}
    for i in implements:
        inherited_fields.update(i.fields_dict)

    self_fields: dict[str, QtGqlFieldDefinition] = {
        name: evaluate_field(type_info, name, field)
        for name, field in obj.fields.items()
        if name not in inherited_fields.keys()
    }

    inherited_fields.update(self_fields)
    return InterfaceOrObjectOptions(
        implements=implements,
        all_fields=inherited_fields,
        unique_fields=tuple(self_fields.values()),
    )


def _evaluate_object_type(
    type_info: SchemaTypeInfo,
    type_: gql_def.GraphQLObjectType,
) -> QtGqlObjectType | None:
    t_name: str = type_.name
    if evaluated := type_info.get_object_type(t_name):
        return evaluated

    options = _evaluate_object_fields(type_info, type_)

    ret = QtGqlObjectType(
        name=t_name,
        interfaces_raw=options.implements,
        docstring=type_.description,
        fields_dict=options.all_fields,
        unique_fields=options.unique_fields,
        is_root=t_name in type_info.root_types_names,
    )
    type_info.add_objecttype(ret)
    for interface in type_.interfaces:
        qtgql_interface = _evaluate_interface_type(type_info, interface)
        qtgql_interface.implementations[type_.name] = ret
    return ret


def _evaluate_interface_type(
    type_info: SchemaTypeInfo,
    interface: gql_def.GraphQLInterfaceType,
) -> QtGqlInterface:
    if ret := type_info.interfaces.get(interface.name, None):
        return ret
    options = _evaluate_object_fields(type_info, interface)
    ret = QtGqlInterface(
        name=interface.name,
        interfaces_raw=options.implements,
        docstring=interface.description,
        fields_dict=options.all_fields,
        unique_fields=options.unique_fields,
    )
    if ret.name == "Node":
        id_field = ret.fields[0]
        if id_field.type != BuiltinScalars.ID or id_field.type.is_optional:
            raise QtGqlException(
                f"Node is a reserved type, id field on `Node` interface must be of type `ID!`"
                f"\n Got: {interface.fields['id']}",
            )
    type_info.interfaces[ret.name] = ret
    return ret


def _evaluate_enum(
    type_info: SchemaTypeInfo,
    enum: gql_def.GraphQLEnumType,
) -> QtGqlEnumDefinition | None:
    name: str = enum.name

    if definition := type_info.enums.get(name, None):
        return definition

    ret = QtGqlEnumDefinition(
        name=name,
        members=[
            EnumValue(name=enum.value, index=index, description=enum.description or "")
            for index, enum in enumerate(enum.values.values())
        ],
    )

    type_info.enums[name] = ret
    return ret


def evaluate_schema(
    schema: gql_def.GraphQLSchema,
    custom_scalars: CustomScalarMap,
) -> SchemaTypeInfo:
    type_info = SchemaTypeInfo(schema, custom_scalars)
    for name, type_ in type_info.schema_definition.type_map.items():
        if name.startswith("__"):
            continue
        if object_definition := is_object_definition(type_):
            if object_type := _evaluate_object_type(type_info, object_definition):
                if object_definition is type_info.schema_definition.query_type:
                    type_info.operation_types[OperationType.QUERY.value] = object_type
                elif object_definition is type_info.schema_definition.mutation_type:
                    type_info.operation_types[OperationType.MUTATION.value] = object_type
                elif object_definition is type_info.schema_definition.subscription_type:
                    type_info.operation_types[OperationType.SUBSCRIPTION.value] = object_type

        elif enum_def := is_enum_definition(type_):
            if enum := _evaluate_enum(type_info, enum_def):
                type_info.enums[enum.name] = enum

    return type_info



from __future__ import annotations

import functools
from typing import TYPE_CHECKING

import attrs
from attr import define

from qtgqlcodegen.core.graphql_ref import (
    SelectionsSet,
    is_fragment_spread_node,
)
from qtgqlcodegen.utils import HashAbleDict, _replace_tuple_item

if TYPE_CHECKING:
    from graphql.language import ast as gql_lang


@define
class _UnwrappedSelectionSet:
    """Selection set without fragments and the fragments that were used."""

    selection_set: SelectionsSet
    used_fragments: dict[str, gql_lang.FragmentDefinitionNode] = attrs.Factory(dict)


@functools.lru_cache(None)
def _unwrap_fragment_spread(
    available_fragments: HashAbleDict[str, gql_lang.FragmentDefinitionNode],
    frag: gql_lang.FragmentDefinitionNode,
) -> tuple[dict[str, gql_lang.FragmentDefinitionNode], SelectionsSet]:
    ret: list[gql_lang.SelectionNode] = []
    used_frags: dict[str, gql_lang.FragmentDefinitionNode] = {}
    for node in frag.selection_set.selections:
        # there might be inner fragments.
        if frag_spread := is_fragment_spread_node(node):
            frag = available_fragments[frag_spread.name.value]
            used_frags[frag.name.value] = frag
            res = _unwrap_fragment_spread(available_fragments, frag)
            used_frags.update(res[0])
            ret.extend(res[1])
        else:
            ret.append(node)

    return used_frags, tuple(ret)


def unwrap_frag_spreads(
    available_fragments: HashAbleDict[str, gql_lang.FragmentDefinitionNode],
    selections: SelectionsSet,
    ret: _UnwrappedSelectionSet | None = None,
) -> _UnwrappedSelectionSet:
    if not ret:
        ret = _UnwrappedSelectionSet(selection_set=selections)
    for i, selection in enumerate(ret.selection_set):
        if frag_spread := is_fragment_spread_node(selection):
            resolved = available_fragments[frag_spread.name.value]
            ret.used_fragments[resolved.name.value] = resolved
            used_by_frag, frag_selections = _unwrap_fragment_spread(available_fragments, resolved)
            ret.selection_set = _replace_tuple_item(ret.selection_set, i, frag_selections)
            ret.used_fragments.update(used_by_frag)
    return ret



from __future__ import annotations

from typing import TYPE_CHECKING

import graphql
from attr import define

from qtgqlcodegen.core.exceptions import QtGqlException
from qtgqlcodegen.core.template import CmakeTemplateContext, cmake_template
from qtgqlcodegen.operation.evaluation import evaluate_operations
from qtgqlcodegen.operation.template import OperationTemplateContext
from qtgqlcodegen.schema.evaluation import evaluate_schema
from qtgqlcodegen.schema.template import (
    SchemaTemplateContext,
    operation_cpp_template,
    operation_hpp_template,
    schema_types_template_hpp,
)
from qtgqlcodegen.types import BuiltinScalars
from qtgqlcodegen.utils import FileSpec

if TYPE_CHECKING:
    from qtgqlcodegen.config import QtGqlConfig


@define
class OperationOutput:
    name: str
    sources: list[FileSpec]


@define
class GenerationOutput:
    schema: FileSpec
    operations: list[OperationOutput]

    def dump(self) -> None:
        self.schema.dump()
        for op in self.operations:
            for source in op.sources:
                source.dump()


class SchemaGenerator:
    def __init__(self, config: QtGqlConfig, schema: graphql.GraphQLSchema):
        self.gql_schema = schema
        self.config = config
        self.schema_type_info = evaluate_schema(schema, self.config.custom_scalars)

    def generate(self) -> GenerationOutput:
        operations = self._generate_operations()
        context = SchemaTemplateContext(
            enums=list(self.schema_type_info.enums.values()),
            types=[
                t
                for name, t in self.schema_type_info.object_types.items()
                if name not in BuiltinScalars.keys
            ],
            interfaces=list(self.schema_type_info.interfaces.values()),
            input_objects=list(self.schema_type_info.input_objects.values()),
            config=self.config,
        )
        schema_hpp = FileSpec(
            content=schema_types_template_hpp(context),
            path=self.config.generated_dir / "schema.hpp",
        )

        return GenerationOutput(
            schema=schema_hpp,
            operations=operations,
        )

    def _generate_operations(self) -> list[OperationOutput]:
        operations_document = graphql.parse(self.config.operations_dir.read_text("utf-8"))
        # validate the operation against the static schema
        if errors := graphql.validate(self.gql_schema, operations_document):
            raise QtGqlException([error.formatted for error in errors])

        operations = evaluate_operations(operations_document, self.schema_type_info)
        ret: list[OperationOutput] = []
        for op_name, op in operations.items():
            context = OperationTemplateContext(
                operation=op,
                config=self.config,
                debug=self.config.debug,
            )

            ret.append(
                OperationOutput(
                    name=op_name,
                    sources=[
                        FileSpec(
                            content=operation_hpp_template(context=context),
                            path=self.config.generated_dir / f"{op_name}.hpp",
                        ),
                        FileSpec(
                            content=operation_cpp_template(context=context),
                            path=self.config.generated_dir / f"{op_name}.cpp",
                        ),
                    ],
                ),
            )

        return ret

    def dump(self):
        generation_output = self.generate()

        cmake = FileSpec(
            content=cmake_template(
                CmakeTemplateContext(config=self.config, generation_output=generation_output),
            ),
            path=self.config.generated_dir / "CMakeLists.txt",
        )
        generation_output.dump()
        cmake.dump()



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from attr import define

from qtgqlcodegen.core.cppref import QtGqlTypes
from qtgqlcodegen.core.template import template_env

if TYPE_CHECKING:
    from qtgqlcodegen.config import QtGqlConfig
    from qtgqlcodegen.operation.definitions import QtGqlOperationDefinition


@define(slots=False)
class OperationTemplateContext:
    operation: QtGqlOperationDefinition
    config: QtGqlConfig
    debug: bool = False
    qtgql_types: ClassVar[type[QtGqlTypes]] = QtGqlTypes

    @property
    def ns(self) -> str:
        return self.operation.name.lower()

    @property
    def schema_ns(self) -> str:
        return self.config.env_name


OPERATION_HPP_TEMPLATE = template_env.get_template("operation.jinja.hpp")
OPERATION_CPP_TEMPLATE = template_env.get_template("operation.jinja.cpp")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Hashable, Literal, TypeVar

from attr import define

if TYPE_CHECKING:
    from pathlib import Path


@define
class FileSpec:
    path: Path
    content: str

    def dump(self) -> None:
        self.path.write_text(self.content, "UTF-8")


class AntiForwardRef:
    """i.e:

    Union["someString"] would return a ForwardRef, this class is a
    simple hack to just return a type contains the name. Also, this is a
    workaround for types that reference each-other, otherwise it would
    cause recursion error.
    """

    name: str
    type_map: dict

    @classmethod
    def resolve(cls) -> Any:
        return cls.type_map[cls.name]


def anti_forward_ref(name: str, type_map: dict) -> type[AntiForwardRef]:
    return type(name, (AntiForwardRef,), {"name": name, "type_map": type_map})


UNSET: Any = Literal["UNSET"]

T = TypeVar("T")


def require(v: T | None) -> T:  # pragma: no cover
    if not v:
        raise AttributeError(f"{v} returned no value")
    return v


def freeze(self, key, value):  # pragma: no cover
    raise PermissionError("setattr called on frozen type")


def _replace_tuple_item(
    original: tuple,
    at: int,
    replace: tuple,
):
    return original[0:at] + replace + original[at + 1 : len(original)]


T_Key = TypeVar("T_Key", bound=Hashable)
T_Value = TypeVar("T_Value")


class HashAbleDict(dict[T_Key, T_Value]):
    """Dict that hashes the keys.

    supports only flat dicts.
    """

    def __hash__(self):
        return hash(frozenset(self.keys()))



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, NamedTuple

from attr import define

if TYPE_CHECKING:
    from typing_extensions import Literal, Self, TypeAlias

CppAccessor: TypeAlias = 'Literal["::"] | Literal["."] | Literal["->"]'


@define(slots=False)
class CppAttribute:
    class InnerCppAttribute(NamedTuple):
        accessor: CppAccessor
        attr: CppAttribute

        @classmethod
        def create(cls, accessor: CppAccessor, attr: str) -> Self:
            return cls(
                accessor=accessor,
                attr=CppAttribute(attr),
            )

    attr: str
    inner: CppAttribute.InnerCppAttribute | None = None

    def ns_add(self, inner: str) -> Self:
        if self.inner:
            self.inner.attr.ns_add(inner)
        else:
            self.inner = CppAttribute.InnerCppAttribute.create(
                "::",
                inner,
            )
        return self

    def build(self) -> str:
        if self.inner:
            return f"{self.attr}{self.inner.accessor}{self.inner.attr.build()}"
        return self.attr

    @cached_property
    def name(self) -> str:
        return self.build()

    @cached_property
    def last(self) -> str:
        if self.inner:
            return self.inner.attr.last
        return self.attr


def QtGqlNs() -> CppAttribute:
    return CppAttribute(
        attr="qtgql",
    )


def QtGqlBasesNs() -> CppAttribute:
    return QtGqlNs().ns_add("bases")


class QtGqlTypes:
    QGraphQLList = QtGqlBasesNs().ns_add("ListModelABC")
    NodeInterfaceABC = QtGqlBasesNs().ns_add("NodeInterfaceABC")
    ObjectTypeABC = QtGqlBasesNs().ns_add("ObjectTypeABC")



from __future__ import annotations

import glob
import importlib.util
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import rich
import typer

import qtgqlcodegen

if TYPE_CHECKING:
    from qtgqlcodegen.config import QtGqlConfig

console = rich.console.Console()
app = typer.Typer(pretty_exceptions_show_locals=False)

QTGQL_CONFIG_FNAME = "qtgqlconfig.py"


def _create_path_link(path: Path) -> str:
    return f"[link={path.resolve()}]{path!s}[/link]"


def _get_config() -> QtGqlConfig:
    console.print("[bold blue]Looking for a config file...")
    res = glob.glob(f"**/{QTGQL_CONFIG_FNAME}", recursive=True)

    if len(res) > 1:
        cwd = Path.cwd()
        results = " \n".join([_create_path_link(cwd / rel) for rel in res])

        console.print(
            f"[bold red]Found more than one config file. {len(res)}" f" found:\n {results}",
        )
        raise typer.Abort()
    elif len(res) < 1:
        console.print(
            f"[bold red]Could not find a config file under {_create_path_link(Path.cwd())}",
        )
        raise typer.Abort()

    mod_path = Path(res[0]).resolve(True)
    spec = importlib.util.spec_from_file_location(QTGQL_CONFIG_FNAME, mod_path)
    assert spec
    module = importlib.util.module_from_spec(spec)
    sys.modules[QTGQL_CONFIG_FNAME] = module
    assert spec.loader
    spec.loader.exec_module(module)
    return module.config


@app.command()
def gen() -> None:
    """Generates types based on your `QtGqlConfig` configuration object."""
    console.print("[bold blue]Generating...")
    with console.status("Still generating...") as s:
        config = _get_config()
        s.update("[green]Configuration file loaded")
        s.update("[bold blue]Just a second I need some coffee ☕")
        config.generate()

    console.print(
        "[bold green]Generated to" f"{_create_path_link(config.generated_dir)} successfully!",
    )


@app.command()
def hotreload():  # pragma: no cover
    raise NotImplementedError


@app.command()
def version() -> None:
    """Show the version of qtgql."""
    console.print(f"[bold blue]{qtgqlcodegen.__version__}")


def entrypoint():  # pragma: no cover
    app()



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, TypeVar

from graphql import language as gql_lang
from graphql.type import definition as gql_def

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

T_AST_Node = TypeVar("T_AST_Node", bound=gql_lang.Node)
SelectionsSet: TypeAlias = "tuple[gql_lang.SelectionNode, ...]"


def ast_identifier_factory(
    expected: type[T_AST_Node],
) -> Callable[[gql_lang.Node], T_AST_Node | None]:
    def type_guarder(node: gql_lang.ast.Node) -> T_AST_Node | None:
        if isinstance(node, expected):
            return node

    return type_guarder


is_selection_set = ast_identifier_factory(gql_lang.ast.SelectionSetNode)
is_inline_fragment = ast_identifier_factory(gql_lang.InlineFragmentNode)
is_fragment_spread_node = ast_identifier_factory(gql_lang.FragmentSpreadNode)
is_fragment_definition_node = ast_identifier_factory(gql_def.FragmentDefinitionNode)
is_operation_def_node = ast_identifier_factory(gql_def.OperationDefinitionNode)
is_field_node = ast_identifier_factory(gql_def.FieldNode)
is_nonnull_node = ast_identifier_factory(gql_lang.NonNullTypeNode)
is_named_type_node = ast_identifier_factory(gql_lang.NamedTypeNode)


T_Definition = TypeVar("T_Definition", bound=gql_def.GraphQLType)


def definition_identifier_factory(
    expected: type[T_Definition],
) -> Callable[[gql_def.GraphQLType], T_Definition | None]:
    def type_guarder(definition: gql_def.GraphQLType) -> T_Definition | None:
        if isinstance(definition, expected):
            return definition

    return type_guarder


is_object_definition = definition_identifier_factory(gql_def.GraphQLObjectType)
is_enum_definition = definition_identifier_factory(gql_def.GraphQLEnumType)
is_list_definition = definition_identifier_factory(gql_def.GraphQLList)
is_scalar_definition = definition_identifier_factory(gql_def.GraphQLScalarType)
is_union_definition = definition_identifier_factory(gql_def.GraphQLUnionType)
is_non_null_definition = definition_identifier_factory(gql_def.GraphQLNonNull)
is_input_definition = definition_identifier_factory(gql_def.GraphQLInputObjectType)
is_interface_definition = definition_identifier_factory(gql_def.GraphQLInterfaceType)


ID_SELECTION_NODE = (
    gql_lang.FieldNode(name=gql_lang.NameNode(value="id"), arguments=(), directives=()),
)

TYPENAME_SELECTION_NODE = (
    gql_lang.FieldNode(name=gql_lang.NameNode(value="__typename"), arguments=(), directives=()),
)


def inject_selection_factory(
    node: tuple[gql_lang.FieldNode],
) -> Callable[[gql_lang.SelectionSetNode], None]:
    def injector(selection_set: gql_lang.SelectionSetNode) -> None:
        selection_set.selections += node

    return injector


inject_id_selection = inject_selection_factory(ID_SELECTION_NODE)
inject_typename_selection = inject_selection_factory(TYPENAME_SELECTION_NODE)


def selection_set_search_factory(
    selection_name: str,
) -> Callable[[gql_lang.SelectionSetNode], bool]:
    def factory(
        selection_set: gql_lang.SelectionSetNode,
    ):
        for node in selection_set.selections:
            if field := is_field_node(node):
                if field.name.value == selection_name:
                    return True
        return False

    return factory


has_id_selection = selection_set_search_factory("id")
has_typename_selection = selection_set_search_factory("__typename")
